==> pftpd-0.30.tar.gz: pftpd-0.30/plib/server.c <==
/*
** server.c - TCP/IP socket server code
**
** Copyright (c) 1997-1999 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#include <stdio.h>
#include <syslog.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#ifdef INET6
#include <netdb.h>
#endif

#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef HAVE_THREADS
#include <sys/wait.h>
#endif

#include "plib/safeio.h"
#include "plib/safestr.h"
#include "plib/support.h"
#include "plib/server.h"



SERVER *
server_init(int fd,
#ifdef INET6
	    char *addr,
	    char *port,
#else
	    UINT32 addr,
	    UINT16 port,
#endif
	    int backlog,
	    int max_clients,
	    void (*handler)(CLIENT *cp))
{
    static int one = 1;
#ifdef INET6
    struct addrinfo hints, *res;
    union sockunion sin;
#else
    struct sockaddr_in sin;
#endif
    socklen_t slen;
    SERVER *sp;
    
    
    sp = s_malloc(sizeof(*sp));

    pthread_attr_init(&sp->ca_detached);
    pthread_attr_setdetachstate(&sp->ca_detached,
				PTHREAD_CREATE_DETACHED);
    
    pthread_mutex_init(&sp->clients_mtx, NULL);
    pthread_cond_init(&sp->clients_cv, NULL);
    sp->clients_cur = 0;
    sp->clients_max = max_clients;
	
    if (fd < 0)
    {
#ifdef INET6
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_INET6;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;
	if (getaddrinfo(addr, port, &hints, &res)) {
	    syslog(LOG_ERR, "getaddrinfo(%s, %s) failed: %m", addr, port);
	    s_free(sp);
	    return NULL;
	}

	sp->fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (sp->fd < 0)
	{
	    syslog(LOG_ERR, "socket(AF_INET6, SOCK_STREAM) failed: %m");
	    s_free(sp);
	    return NULL;
	}

	(void) setsockopt(sp->fd, SOL_SOCKET, SO_REUSEADDR,
			  (void *) &one, sizeof(one));
	
	if (s_bind(sp->fd, res->ai_addr, res->ai_addrlen) < 0)
	{
	    char buf1[NI_MAXHOST];


	    getnameinfo((struct sockaddr *)&sp->sin, SA_LEN(&sp->sin),
			buf1, sizeof(buf1), NULL, 0, NI_NUMERICHOST);
	    syslog(LOG_ERR, "bind(%d,%s,%d) failed: %m", sp->fd, buf1, port);
	    
	    s_close(sp->fd);
	    return NULL;
	}
	memcpy(&sp->sin, res->ai_addr, res->ai_addrlen);
	freeaddrinfo(res);
#else
	sp->fd = socket(AF_INET, SOCK_STREAM, 0);
	if (sp->fd < 0)
	{
	    syslog(LOG_ERR, "socket(AF_INET, SOCK_STREAM) failed: %m");
	    s_free(sp);
	    return NULL;
	}

	(void) setsockopt(sp->fd, SOL_SOCKET, SO_REUSEADDR,
			  (void *) &one, sizeof(one));
	
	memset(&sp->sin, 0, sizeof(sin));
	sp->sin.sin_family = AF_INET;
	sp->sin.sin_addr.s_addr = htonl(addr);
	sp->sin.sin_port = htons(port);

	if (s_bind(sp->fd, (struct sockaddr *) &sp->sin, sizeof(sp->sin)) < 0)
	{
	    char buf1[16];

	    
	    syslog(LOG_ERR, "bind(%d,%s,%d) failed: %m",
		   sp->fd,
		   s_inet_ntoa(sp->sin.sin_addr, buf1, sizeof(buf1)),
		   port);
	    
	    s_close(sp->fd);
	    return NULL;
	}
#endif

	if (listen(sp->fd, backlog) < 0)
	{
	    syslog(LOG_ERR, "listen(%d, %d) failed: %m", sp->fd, backlog);
	    s_close(sp->fd);
	    return NULL;
	}
    }
    else
    {
	sp->fd = fd;
	slen = sizeof(sp->sin);
	getsockname(sp->fd, (struct sockaddr *) &sp->sin, &slen);
    }

    sp->handler = handler;
    
    return sp;
}



static void *
client_thread(void *vp)
{
    CLIENT *cp = (CLIENT *) vp;
    SERVER *sp = cp->sp;
    
    sp->handler(cp);
    
    s_close(cp->fd);
    
#ifdef HAVE_THREADS
    if (sp->clients_max > 0)
    {
	pthread_mutex_lock(&sp->clients_mtx);
	if (sp->clients_cur == sp->clients_max)
	    pthread_cond_signal(&sp->clients_cv);
	
	sp->clients_cur--;
	pthread_mutex_unlock(&sp->clients_mtx);
    }
#endif

    s_free(cp);
    return NULL;
}

int
server_run_one(SERVER *sp,
	       int client_fd,
	       int nofork)
{
    CLIENT *cp;
    pthread_t tid;
#ifndef HAVE_THREADS
    pid_t pid;
    int status;
#endif
    int ecode;
    socklen_t slen;
    
    
    cp = s_malloc(sizeof(*cp));
    cp->fd = client_fd;
    cp->sp = sp;
    
    slen = sizeof(cp->rsin);
    if (getpeername(cp->fd,
		    (struct sockaddr *) &cp->rsin, &slen) < 0)
    {
	syslog(LOG_ERR, "getpeername(%d): %m", cp->fd);
	return EXIT_FAILURE;
    }
    
    slen = sizeof(cp->lsin);
    if (getsockname(cp->fd,
		    (struct sockaddr *) &cp->lsin, &slen) < 0)
    {
	syslog(LOG_ERR, "getsockname(%d): %m", cp->fd);
	return EXIT_FAILURE;
    }
    
    if (nofork)
    {
	(void) client_thread(cp);
	return EXIT_SUCCESS;
    }
    else
    {
#ifdef HAVE_THREADS
	if (sp->clients_max > 0)
	{
	    /* XXX: Return error */
	    
	    pthread_mutex_lock(&sp->clients_mtx);
	    while (sp->clients_cur >= sp->clients_max)
		pthread_cond_wait(&sp->clients_cv, &sp->clients_mtx);
	    
	    sp->clients_cur++;
	    pthread_mutex_unlock(&sp->clients_mtx);
	}

	ecode = pthread_create(&tid,
			       &sp->ca_detached,
			       client_thread, (void *) cp);
	if (ecode)
	{
	    syslog(LOG_ERR, "pthread_create(handler-thread) failed: %s",
		   strerror(ecode));
	    return EXIT_FAILURE;
	}
#else
	
	/* Try to reap the status of as many subprocesses as possible */
	
	/*
	** XXX: This will break if we are using multiple
	**      SERVER's in a single process and isn't using
	**      threads.
	*/
	while (sp->clients_cur > 0 &&
	       (pid = waitpid((pid_t) -1, &status, WNOHANG)) > 0)
	{
	    if (WIFEXITED(status) || WIFSIGNALED(status))
	    {
		sp->clients_cur--;
	    }
	}

	if (sp->clients_max > 0)
	{
	    /* Wait for atleast one slot to be available */
	    while (sp->clients_cur >= sp->clients_max &&
		   (pid = waitpid((pid_t) -1, &status, 0)) > 0)
	    {
		if (WIFEXITED(status) || WIFSIGNALED(status))
		{
		    sp->clients_cur--;
		}
	    }
	}
	
	while ((status = fork()) < 0 && errno == EAGAIN)
	{
	    /* Fork failed - too many processes */
	    sleep(1);

	    pid = waitpid((pid_t) -1, &status,
			  (sp->clients_cur > 0) ? 0 : WNOHANG);
		
	    if (pid > 0 && (WIFEXITED(status) || WIFSIGNALED(status)))
	    {
		sp->clients_cur--;
	    }
	}
	      
	if (status < 0)
	{
	    syslog(LOG_ERR, "fork() failed: %m");
	    s_abort();
	}
	
	if (status == 0)
	{
	    /* In child process */
	    (void) client_thread(cp);

	    _exit(EXIT_SUCCESS);
	}

	sp->clients_cur++;

	s_close(cp->fd);
	s_free(cp);
#endif
    }

    return EXIT_SUCCESS;
}



int
server_run_all(SERVER *sp)
{
    int fd;

    
    while ((fd = s_accept(sp->fd, NULL, NULL)) >= 0)
	server_run_one(sp, fd, 0);

    syslog(LOG_ERR, "accept() failed: %m");
    return -1;
}

==> pftpd-0.37.tar.gz: pftpd-0.37/plib/server.c <==
/*
** server.c - TCP/IP socket server code
**
** Copyright (c) 1997-2000 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "plib/config.h"

#include <stdio.h>
#include <syslog.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef HAVE_THREADS
#include <sys/wait.h>
#endif

#include "plib/safeio.h"
#include "plib/safestr.h"
#include "plib/support.h"
#include "plib/aalloc.h"
#include "plib/sockaddr.h"

#include "plib/server.h"



SERVER *
server_init(int fd,
	    const struct sockaddr_gen *sg,
	    int backlog,
	    int max_clients,
	    void (*handler)(CLIENT *cp))
{
    static int one = 1;
    socklen_t slen;
    SERVER *sp;
    
    
    A_NEW(sp);

    pthread_attr_init(&sp->ca_detached);
    pthread_attr_setdetachstate(&sp->ca_detached,
				PTHREAD_CREATE_DETACHED);
    
    pthread_mutex_init(&sp->clients_mtx, NULL);
    pthread_cond_init(&sp->clients_cv, NULL);
    sp->clients_cur = 0;
    sp->clients_max = max_clients;
	
    if (fd < 0)
    {
	if (sg == NULL)
	{
	    syslog(LOG_ERR,
		   "server_init: NULL address and no file descriptor");
	    a_free(sp);
	    return NULL;
	}
	
	sp->sin = *sg;
	
	sp->fd = socket(SGFAM(sp->sin), SOCK_STREAM, 0);
#ifdef HAVE_IPV6
	if (sp->fd < 0 &&
	    (errno == EAFNOSUPPORT || errno == EPFNOSUPPORT) &&
	    SGFAM(sp->sin) == AF_INET6)
	{
#if 0
	    /* Try to convert to IPv4 format... */
	    struct in6_addr *addr6 = (struct in6_addr *) SGADDRP(sp->sin);
	    
	    if (IN6_IS_ADDR_V4MAPPED(addr6))
	    {
		UINT32 addr4 = addr6->s6_addr32[3];
		
		SGFAM(sp->sin) = AF_INET;
		* (UINT32 *) SGADDRP(sp->sin) = addr4;
	    }
#endif

	    /* Let's try with an IPv4 socket - who knows, it might work */
	    errno = 0;
	    sp->fd = socket(AF_INET, SOCK_STREAM, 0);
	}
#endif
	
	if (sp->fd < 0)
	{
	    syslog(LOG_ERR, "socket(%s, SOCK_STREAM) failed (errno=%d): %m",
		   (SGFAM(sp->sin) == AF_INET ? "AF_INET" : "AF_INET6"),
		   errno);
	    a_free(sp);
	    return NULL;
	}

	(void) setsockopt(sp->fd, SOL_SOCKET, SO_REUSEADDR,
			  (void *) &one, sizeof(one));

	if (s_bind(sp->fd, (struct sockaddr *) &sp->sin,
		   SGSOCKSIZE(sp->sin)) < 0)
	{
	    char buf1[16];

	    syslog(LOG_ERR, "bind(%d,%s) failed: %m",
		   sp->fd,
		   s_inet_ntox(&sp->sin, buf1, sizeof(buf1)));
	    
	    s_close(sp->fd);
	    return NULL;
	}
    }
    else
    {
	sp->fd = fd;
	slen = sizeof(sp->sin);
	getsockname(sp->fd, (struct sockaddr *) &sp->sin, &slen);
    }

    /* We do this outside the 'if' clause to support broken
       Inetd implementations */
    
    if (listen(sp->fd, backlog) < 0)
    {
	syslog(LOG_ERR, "listen(%d, %d) failed: %m", sp->fd, backlog);
	s_close(sp->fd);
	return NULL;
    }
	
    sp->handler = handler;
    
    return sp;
}



static void *
client_thread(void *vp)
{
    CLIENT *cp = (CLIENT *) vp;
    SERVER *sp = cp->sp;
    
    sp->handler(cp);
    
    s_close(cp->fd);
    
#ifdef HAVE_THREADS
    if (sp->clients_max > 0)
    {
	pthread_mutex_lock(&sp->clients_mtx);
	if (sp->clients_cur == sp->clients_max)
	    pthread_cond_signal(&sp->clients_cv);
	
	sp->clients_cur--;
	pthread_mutex_unlock(&sp->clients_mtx);
    }
#endif

    a_free(cp);
    return NULL;
}

int
server_run_one(SERVER *sp,
	       int client_fd,
	       int nofork)
{
    CLIENT *cp;
    pthread_t tid;
#ifndef HAVE_THREADS
    pid_t pid;
    int status;
#endif
    int ecode;
    socklen_t slen;
    
    
    A_NEW(cp);
    cp->fd = client_fd;
    cp->sp = sp;
    
    slen = sizeof(cp->rsin);
    if (getpeername(cp->fd,
		    (struct sockaddr *) &cp->rsin, &slen) < 0)
    {
	syslog(LOG_ERR, "getpeername(%d): %m", cp->fd);
	return EXIT_FAILURE;
    }
    
    slen = sizeof(cp->lsin);
    if (getsockname(cp->fd,
		    (struct sockaddr *) &cp->lsin, &slen) < 0)
    {
	syslog(LOG_ERR, "getsockname(%d): %m", cp->fd);
	return EXIT_FAILURE;
    }
    
    if (nofork)
    {
	(void) client_thread(cp);
	return EXIT_SUCCESS;
    }
    else
    {
#ifdef HAVE_THREADS
	if (sp->clients_max > 0)
	{
	    /* XXX: Return error */
	    
	    pthread_mutex_lock(&sp->clients_mtx);
	    while (sp->clients_cur >= sp->clients_max)
		pthread_cond_wait(&sp->clients_cv, &sp->clients_mtx);
	    
	    sp->clients_cur++;
	    pthread_mutex_unlock(&sp->clients_mtx);
	}

	ecode = pthread_create(&tid,
			       &sp->ca_detached,
			       client_thread, (void *) cp);
	if (ecode)
	{
	    syslog(LOG_ERR, "pthread_create(handler-thread) failed: %s",
		   strerror(ecode));
	    return EXIT_FAILURE;
	}

#else
	
	/* Try to reap the status of as many subprocesses as possible */
	
	/*
	** XXX: This will break if we are using multiple
	**      SERVER's in a single process and isn't using
	**      threads.
	*/
	while (sp->clients_cur > 0 &&
	       (pid = waitpid((pid_t) -1, &status, WNOHANG)) > 0)
	{
	    if (WIFEXITED(status) || WIFSIGNALED(status))
	    {
		sp->clients_cur--;
	    }
	}

	if (sp->clients_max > 0)
	{
	    /* Wait for atleast one slot to be available */
	    while (sp->clients_cur >= sp->clients_max &&
		   (pid = waitpid((pid_t) -1, &status, 0)) > 0)
	    {
		if (WIFEXITED(status) || WIFSIGNALED(status))
		{
		    sp->clients_cur--;
		}
	    }
	}
	
	while ((status = fork()) < 0 && errno == EAGAIN)
	{
	    /* Fork failed - too many processes */
	    sleep(1);

	    pid = waitpid((pid_t) -1, &status,
			  (sp->clients_cur > 0) ? 0 : WNOHANG);
		
	    if (pid > 0 && (WIFEXITED(status) || WIFSIGNALED(status)))
	    {
		sp->clients_cur--;
	    }
	}
	      
	if (status < 0)
	{
	    syslog(LOG_ERR, "fork() failed: %m");
	    s_abort();
	}
	
	if (status == 0)
	{
	    /* In child process */
	    (void) client_thread(cp);

	    _exit(EXIT_SUCCESS);
	}

	sp->clients_cur++;

	s_close(cp->fd);
	a_free(cp);
#endif
    }

    return EXIT_SUCCESS;
}



int
server_run_all(SERVER *sp)
{
    int fd;


    while (sp->fd >= 0)
    {
	fd = s_accept(sp->fd, NULL, NULL);
	
	if (fd >= 0)
	    server_run_one(sp, fd, 0);
	else
	{
	    syslog(LOG_ERR, "accept() failed: %m");
	    
	    switch (errno)
	    {
	      case EBADF:
	      case EMFILE:
	      case ENODEV:
	      case ENOMEM:
	      case ENOTSOCK:
	      case EOPNOTSUPP:
	      case EWOULDBLOCK:
	      case EINVAL:
	      case ENFILE:
		return -1;
	    }
	}
    }

    return -1;
}
